__title__ = "cycletls"
__description__ = "Python HTTP for Humans."
__url__ = "https://cycletls.readthedocs.io"
__version__ = "0.0.3"
__build__ = 0x022701
__author__ = "Danny Dasilva"
__author_email__ = "dannydasilva.solutions@gmail.com"
__license__ = " GNU General Public License v3 (GPLv3)"
__copyright__ = "Copyright 2022 Danny Dasilva"
__cake__ = "\u2728 \U0001f370 \u2728"
